unit Tests;

interface

uses
  Classes, SysUtils, TestCase;

function GetTests: TTestCases;


implementation

uses
  Parser, Executor, ParserPascal;

type

  { TTestRun }

  TTestRun = class(TTestCase)
  private
    Output: string;
    Error: string;
    procedure ExecutorOutput(Text: string);
    procedure InterpreterError(Pos: TPoint; Text: string);
  public
    Source: TStringList;
    ExpectedOutput: string;
    procedure Run; override;
    constructor Create; override;
    destructor Destroy; override;
  end;

function GetTests: TTestCases;
begin
  Result := TTestCases.Create;
  with TTestRun(Result.AddNew('Basic program', TTestRun)) do begin
    Source.Add('program Test;');
    Source.Add('begin');
    Source.Add('end.');
    ExpectedOutput := '';
  end;
  with TTestRun(Result.AddNew('WriteLn', TTestRun)) do begin
    Source.Add('begin');
    Source.Add('  WriteLn(''Test'');');
    Source.Add('end.');
    ExpectedOutput := 'Test' + LineEnding;
  end;
  with TTestRun(Result.AddNew('Write', TTestRun)) do begin
    Source.Add('begin');
    Source.Add('  Write(''Test'');');
    Source.Add('end.');
    ExpectedOutput := 'Test';
  end;
  with TTestRun(Result.AddNew('Integer constants summation', TTestRun)) do begin
    Source.Add('begin');
    Source.Add('  WriteLn(IntToStr(1 + 2));');
    Source.Add('end.');
    ExpectedOutput := '3' + LineEnding;
  end;
  with TTestRun(Result.AddNew('String concatenation', TTestRun)) do begin
    Source.Add('begin');
    Source.Add('  WriteLn(''Test'' + ''Now'');');
    Source.Add('end.');
    ExpectedOutput := 'TestNow' + LineEnding;
  end;
  with TTestRun(Result.AddNew('String variable assignment', TTestRun)) do begin
    Source.Add('var');
    Source.Add('  A: string;');
    Source.Add('begin');
    Source.Add('  A := ''Test'';');
    Source.Add('  WriteLn(A);');
    Source.Add('end.');
    ExpectedOutput := 'Test' + LineEnding;
  end;
  with TTestRun(Result.AddNew('String constant', TTestRun)) do begin
    Source.Add('const');
    Source.Add('  A: string = ''Test'';');
    Source.Add('begin');
    Source.Add('  WriteLn(A);');
    Source.Add('end.');
    ExpectedOutput := 'Test' + LineEnding;
  end;
  with TTestRun(Result.AddNew('String with single quote', TTestRun)) do begin
    Source.Add('begin');
    Source.Add('  Write(''One''''Two'');');
    Source.Add('end.');
    ExpectedOutput := 'One''Two';
  end;
  with TTestRun(Result.AddNew('if-then true', TTestRun)) do begin
    Source.Add('begin');
    Source.Add('  if 1 = 1 then WriteLn(''Yes'');');
    Source.Add('end.');
    ExpectedOutput := 'Yes' + LineEnding;
  end;
  with TTestRun(Result.AddNew('if-then false', TTestRun)) do begin
    Source.Add('begin');
    Source.Add('  if 1 = 2 then WriteLn(''Yes'');');
    Source.Add('end.');
    ExpectedOutput := '';
  end;
  with TTestRun(Result.AddNew('if-then-else true', TTestRun)) do begin
    Source.Add('begin');
    Source.Add('  if 1 = 1 then WriteLn(''Yes'') else WriteLn(''No'');');
    Source.Add('end.');
    ExpectedOutput := 'Yes' + LineEnding;
  end;
  with TTestRun(Result.AddNew('if-then-else false', TTestRun)) do begin
    Source.Add('begin');
    Source.Add('  if 1 = 2 then WriteLn(''Yes'') else WriteLn(''No'');');
    Source.Add('end.');
    ExpectedOutput := 'No' + LineEnding;
  end;
  with TTestRun(Result.AddNew('for-to-do', TTestRun)) do begin
    Source.Add('var');
    Source.Add('  I: Integer;');
    Source.Add('begin');
    Source.Add('  for I := 0 to 2 do begin');
    Source.Add('    WriteLn(IntToStr(I));');
    Source.Add('  end;');
    Source.Add('end.');
    ExpectedOutput := '0' + LineEnding + '1' + LineEnding + '2' + LineEnding;
  end;
  with TTestRun(Result.AddNew('while-do', TTestRun)) do begin
    Source.Add('var');
    Source.Add('  I: Integer;');
    Source.Add('begin');
    Source.Add('  I := 0;');
    Source.Add('  while I < 3 do begin');
    Source.Add('    WriteLn(IntToStr(I));');
    Source.Add('    I := I + 1;');
    Source.Add('  end;');
    Source.Add('end.');
    ExpectedOutput := '0' + LineEnding + '1' + LineEnding + '2' + LineEnding;
  end;
  with TTestRun(Result.AddNew('repeat-until', TTestRun)) do begin
    Source.Add('var');
    Source.Add('  I: Integer;');
    Source.Add('begin');
    Source.Add('  I := 0;');
    Source.Add('  repeat');
    Source.Add('    WriteLn(IntToStr(I));');
    Source.Add('    I := I + 1;');
    Source.Add('  until I > 2;');
    Source.Add('end.');
    ExpectedOutput := '0' + LineEnding + '1' + LineEnding + '2' + LineEnding;
  end;
  with TTestRun(Result.AddNew('function', TTestRun)) do begin
    Source.Add('function IsZero(A: Integer): Boolean;');
    Source.Add('begin');
    Source.Add('  Result := A = 0;');
    Source.Add('end;');
    Source.Add('');
    Source.Add('begin');
    Source.Add('  WriteLn(BoolToStr(IsZero(0)));');
    Source.Add('  WriteLn(BoolToStr(IsZero(1)));');
    Source.Add('end.');
    ExpectedOutput := '-1' + LineEnding + '0' + LineEnding;
  end;
  with TTestRun(Result.AddNew('function without result usage', TTestRun)) do begin
    Source.Add('function IsZero(A: Integer): Boolean;');
    Source.Add('begin');
    Source.Add('  Result := A = 0;');
    Source.Add('end;');
    Source.Add('');
    Source.Add('begin');
    Source.Add('  IsZero(0);');
    Source.Add('end.');
    ExpectedOutput := '';
  end;
  with TTestRun(Result.AddNew('function var parameter', TTestRun)) do begin
    Source.Add('function Test(var A: Integer): Boolean;');
    Source.Add('begin');
    Source.Add('  A := 10;');
    Source.Add('  Result := 1 = 1;');
    Source.Add('end;');
    Source.Add('');
    Source.Add('var');
    Source.Add('  B: Integer;');
    Source.Add('  C: Boolean;');
    Source.Add('begin');
    Source.Add('  B := 1;');
    Source.Add('  C := Test(B);');
    Source.Add('  WriteLn(IntToStr(B));');
    Source.Add('end.');
    ExpectedOutput := '10' + LineEnding;
  end;
  with TTestRun(Result.AddNew('function var', TTestRun)) do begin
    Source.Add('function Test(A: string): string;');
    Source.Add('var');
    Source.Add('  C: string;');
    Source.Add('begin');
    Source.Add('  C := A;');
    Source.Add('  Result := C;');
    Source.Add('end;');
    Source.Add('');
    Source.Add('begin');
    Source.Add('  WriteLn(Test(''X''));');
    Source.Add('end.');
    ExpectedOutput := 'X' + LineEnding;
  end;
  with TTestRun(Result.AddNew('function const', TTestRun)) do begin
    Source.Add('function Test: string;');
    Source.Add('const');
    Source.Add('  D: string = ''X'';');
    Source.Add('begin');
    Source.Add('  Result := D;');
    Source.Add('end;');
    Source.Add('');
    Source.Add('begin');
    Source.Add('  WriteLn(Test);');
    Source.Add('end.');
    ExpectedOutput := 'X' + LineEnding;
  end;
  with TTestRun(Result.AddNew('procedure', TTestRun)) do begin
    Source.Add('procedure Print(Text: string);');
    Source.Add('begin');
    Source.Add('  WriteLn(Text);');
    Source.Add('end;');
    Source.Add('');
    Source.Add('begin');
    Source.Add('  Print(''Test'');');
    Source.Add('end.');
    ExpectedOutput := 'Test' + LineEnding;
  end;
  with TTestRun(Result.AddNew('procedure var parameter', TTestRun)) do begin
    Source.Add('procedure Test(var A: Integer);');
    Source.Add('begin');
    Source.Add('  A := 10;');
    Source.Add('end;');
    Source.Add('');
    Source.Add('var');
    Source.Add('  B: Integer;');
    Source.Add('begin');
    Source.Add('  B := 1;');
    Source.Add('  Test(B);');
    Source.Add('  WriteLn(IntToStr(B));');
    Source.Add('end.');
    ExpectedOutput := '10' + LineEnding;
  end;
  with TTestRun(Result.AddNew('Single line comment', TTestRun)) do begin
    Source.Add('begin');
    Source.Add('  // WriteLn(''Test'');');
    Source.Add('end.');
    ExpectedOutput := '';
  end;
end;

{ TTestBasic }

procedure TTestRun.ExecutorOutput(Text: string);
begin
  Output := Output + Text;
end;

procedure TTestRun.InterpreterError(Pos: TPoint; Text: string);
begin
  Error := Error + '[' + IntToStr(Pos.X) + ', ' + IntToStr(Pos.Y) + '] ' + Text + LineEnding;
end;

procedure TTestRun.Run;
var
  Parser: TParser;
  Executor: TExecutor;
begin
  Error := '';
  Output := '';
  Parser := TParserPascal.Create;
  try
    Parser.OnError := InterpreterError;
    Parser.Source := Source.Text;
    Parser.Parse;
    if Assigned(Parser.Prog) then
    try
      Executor := TExecutor.Create;
      Executor.Prog := Parser.Prog;
      Executor.OnOutput := ExecutorOutput;
      Executor.Run;
      Evaluate((ExpectedOutput = Output) and (Error = ''));
    finally
      Executor.Free;
    end else TestResult := trFailed;
  finally
    Parser.Prog.Free;
    Parser.Free;
  end;
  Log := 'Source: ' + LineEnding + Source.Text + LineEnding +
    'Expected: ' + LineEnding + ExpectedOutput + LineEnding +
    'Returned: ' + LineEnding + Output + LineEnding +
    'Errors: ' + LineEnding + Error + LineEnding;
  inherited;
end;

constructor TTestRun.Create;
begin
  inherited;
  Source := TStringList.Create;
end;

destructor TTestRun.Destroy;
begin
  FreeAndNil(Source);
  inherited;
end;

end.

